from difflib import ndiff
from io import StringIO
from textwrap import dedent
import os
import re
import unittest

from pyrsistent import v

from virtue import runner
from virtue.reporters import (
    ComponentizedReporter,
    Counter,
    Outputter,
    Recorder,
)


class TestRun(unittest.TestCase):
    def test_it_runs_tests(self):
        result = runner.run(tests=["virtue.tests.samples.one_successful_test"])
        self.assertEqual(result, Counter(successes=1))

    def test_it_runs_unsuccessful_tests(self):
        result = runner.run(
            tests=["virtue.tests.samples.one_unsuccessful_test"],
        )
        self.assertEqual(result, Counter(failures=1))

    def test_it_runs_expected_failing_tests(self):
        result = runner.run(
            tests=["virtue.tests.samples.one_expected_failure"],
        )
        self.assertEqual(result, Counter(expected_failures=1))

    def test_it_runs_unexpectedly_passing_expected_failing_tests(self):
        result = runner.run(
            tests=[
                "virtue.tests.samples.one_expected_failure_mispassing",
            ],
        )
        self.assertEqual(result, Counter(unexpected_successes=1))

    def test_it_runs_subtests(self):
        result = runner.run(tests=["virtue.tests.samples.subtests"])
        self.assertEqual(
            result,
            Counter(
                subtest_failures=1,
                subtest_successes=4,
                successes=3,
                errors=1,
            ),
        )

    def test_it_can_stop_short(self):
        """
        Ooo, I stopped short.
        """

        result = runner.run(
            tests=["virtue.tests.samples.two_unsuccessful_tests"],
            stop_after=1,
        )
        self.assertEqual(
            result,
            Counter(failures=1, successes=result.successes),
        )

    def test_it_can_stop_short_combined_with_errors(self):
        counter = runner.run(
            tests=["virtue.tests.samples.failures_and_errors"],
        )
        self.assertGreater(counter.failures + counter.errors, 3)

        result = runner.run(
            tests=["virtue.tests.samples.failures_and_errors"],
            stop_after=3,
        )
        self.assertEqual(result.failures + result.errors, 3)

    def test_it_can_stop_short_combined_with_unexpected_passing_tests(self):
        """
        RIP Jerry.
        """
        counter = runner.run(
            tests=["virtue.tests.samples.failures_and_unexpected_passes"],
        )
        self.assertGreater(counter.failures + counter.unexpected_successes, 2)

        result = runner.run(
            tests=["virtue.tests.samples.failures_and_unexpected_passes"],
            stop_after=2,
        )
        self.assertEqual(result.failures + result.unexpected_successes, 2)

    def test_warnings_become_errors_by_default(self):
        result = runner.run(
            tests=["virtue.tests.samples.success_and_warning"],
        )
        self.assertEqual(result, Counter(errors=1, successes=1))

    def test_it_runs_tests_by_path_if_you_insist(self):
        import virtue.tests.samples

        path = os.path.join(
            os.path.dirname(virtue.tests.samples.__file__),
            "one_unsuccessful_test.py",
        )
        result = runner.run(tests=[path])
        self.assertEqual(result, Counter(failures=1))

    def test_it_errors_for_paths_that_do_not_exist(self):
        path = os.path.join(
            os.path.dirname(__file__),
            "this_is_a_path_that_doesnt_exist_yes_it_goes_on_and_on_my_friend",
        )
        with self.assertRaises(FileNotFoundError):
            runner.run(tests=[path])

    def test_unittest_TestResult(self):
        result = unittest.TestResult()
        runner.run(
            tests=["virtue.tests.samples.one_successful_test"],
            reporter=result,
        )
        self.assertEqual(result.testsRun, 1)

    def test_Recorder(self):
        result = Recorder()
        runner.run(
            tests=["virtue.tests.samples.one_successful_test"],
            reporter=result,
        )
        import virtue.tests.samples.one_successful_test

        self.assertEqual(
            result,
            Recorder(
                successes=v(
                    virtue.tests.samples.one_successful_test.Foo("test_foo"),
                ),
            ),
        )


class TestRunOutput(unittest.TestCase):
    def assertOutputIs(self, expected, **kwargs):
        reporter = ComponentizedReporter(
            outputter=Outputter(colored=False, line_width=40),
            stream=StringIO(),
            time=lambda: 0,
        )
        runner.run(reporter=reporter, **kwargs)
        got = reporter.stream.getvalue()

        dedented = dedent(expected)
        if dedented.startswith("\n"):
            dedented = dedented[1:]

        # I'm so sorry.
        globbed = re.escape(dedented).replace(re.escape("•"), ".*")

        if re.search(globbed, got, re.DOTALL) is None:
            self.fail(
                "\n    "
                + "\n    ".join(
                    ndiff(got.splitlines(), dedented.splitlines()),
                ),
            )

    def test_unsuccessful_run(self):
        self.assertOutputIs(
            tests=[
                "virtue.tests.samples.one_successful_test",
                "virtue.tests.samples.two_unsuccessful_tests",
            ],
            expected="""
            virtue.tests.samples.one_successful_test
              Foo
                test_foo ...                    [OK]
            virtue.tests.samples.two_unsuccessful_tests
              Bar
                test_bar ...                    [OK]
                test_foo ...                  [FAIL]
              Foo
                test_bar ...                  [FAIL]
                test_foo ...                    [OK]

            ========================================
            [FAIL]
            Traceback (most recent call last):•
              File "•unittest/case.py•", line •, in •
                •
              File "•virtue/tests/samples/two_unsuccessful_tests.py•", line 14, in test_foo
                self.fail("I fail too.")
              •
            AssertionError: I fail too.

            virtue.tests.samples.two_unsuccessful_tests.Bar.test_foo
            ========================================
            [FAIL]
            Traceback (most recent call last):•
              File "•unittest/case.py•", line •, in •
                •
              File "•virtue/tests/samples/two_unsuccessful_tests.py•", line 9, in test_bar
                self.fail("I fail.")
              •
            AssertionError: I fail.

            virtue.tests.samples.two_unsuccessful_tests.Foo.test_bar
            ----------------------------------------
            Ran 5 tests in 0.000s

            FAILED (successes=3, failures=2)
            """,  # noqa: E501
        )

    def test_single_test(self):
        self.assertOutputIs(
            tests=["virtue.tests.samples.one_successful_test"],
            expected="""
            virtue.tests.samples.one_successful_test
              Foo
                test_foo ...                    [OK]

            ----------------------------------------
            Ran 1 test in 0.000s

            PASSED (successes=1)
            """,
        )

    def test_empty_run(self):
        self.assertOutputIs(
            tests=[],
            expected="""

            ----------------------------------------
            Ran 0 tests in 0.000s

            PASSED
            """,
        )

    def test_expected_failure(self):
        self.assertOutputIs(
            tests=["virtue.tests.samples.one_expected_failure"],
            expected="""
            virtue.tests.samples.one_expected_failure
              Foo
                test_foo ...                 [XFAIL]

            ========================================
            [XFAIL]

            virtue.tests.samples.one_expected_failure.Foo.test_foo
            ----------------------------------------
            Ran 1 test in 0.000s

            PASSED (expected_failures=1)
            """,
        )

    def test_unexpected_success(self):
        self.assertOutputIs(
            tests=["virtue.tests.samples.failures_and_unexpected_passes"],
            expected="""
            virtue.tests.samples.failures_and_unexpected_passes
              Foo
                test_bar ...    [UNEXPECTED SUCCESS]
                test_baz ...    [UNEXPECTED SUCCESS]
                test_foo ...                  [FAIL]
                test_quux ...   [UNEXPECTED SUCCESS]
                test_spam ...                   [OK]

            ========================================
            [FAIL]
            Traceback (most recent call last):
              File "•unittest/case.py•", line •, in •
                •
              File "•/failures_and_unexpected_passes.py•", line •, in test_foo
                self.fail("Nope!")
              •
            AssertionError: Nope!

            virtue.tests.samples.failures_and_unexpected_passes.Foo.test_foo
            ========================================
            [UNEXPECTED SUCCESS]

            virtue.tests.samples.failures_and_unexpected_passes.Foo.test_bar
            virtue.tests.samples.failures_and_unexpected_passes.Foo.test_baz
            virtue.tests.samples.failures_and_unexpected_passes.Foo.test_quux
            ----------------------------------------
            Ran 5 tests in 0.000s

            FAILED (successes=1, failures=1, unexpected_successes=3)
            """,
        )

    def test_subtests(self):
        self.assertOutputIs(
            tests=["virtue.tests.samples.subtests"],
            expected="""
            virtue.tests.samples.subtests
              Baz
                test_passing_nested_subtest ... [OK]
                test_passing_subtest ...        [OK]
              Foo
                test_no_subtests ...            [OK]
                test_subtests_one_fail_one_error
                  i=1 ...                     [FAIL]
                  i=3 ...                    [ERROR]

            ========================================
            [FAIL]
            Traceback (most recent call last):
              File "•unittest/case.py•", line •, in •
                •
              File "•/subtests.py•", line •, in test_subtests_one_fail_one_error
                self.fail("Fail!")
              •
            AssertionError: Fail!

            virtue.tests.samples.subtests.Foo.test_subtests_one_fail_one_error (i=1)
            ========================================
            [ERROR]
            Traceback (most recent call last):
              File "•unittest/case.py•", line •, in •
                •
              File "•/subtests.py•", line •, in test_subtests_one_fail_one_error
                raise ZeroDivisionError()•
            ZeroDivisionError•

            virtue.tests.samples.subtests.Foo.test_subtests_one_fail_one_error (i=3)
            ----------------------------------------
            Ran 4 tests with 7 subtests in 0.000s

            FAILED (successes=3, subtest_failures=1, subtest_errors=1)
            """,  # noqa: E501
        )

    def test_output_is_compressed(self):
        """
        Albeit not sorted, so still not terribly reliably :/.
        """
        self.assertOutputIs(
            tests=["virtue.tests.samples.repeated_similar_output"],
            expected="""
            virtue.tests.samples.repeated_similar_output
              Foo
                test_a ...                 [SKIPPED]
                test_b ...                 [SKIPPED]
                test_c ...                 [SKIPPED]

            ========================================
            [SKIPPED]
            Skipped!
            virtue.tests.samples.repeated_similar_output.Foo.test_a
            ========================================
            [SKIPPED]
            Skipped 2!
            virtue.tests.samples.repeated_similar_output.Foo.test_b
            virtue.tests.samples.repeated_similar_output.Foo.test_c
            ----------------------------------------
            Ran 3 tests in 0.000s

            PASSED (skips=3)
            """,
        )



import importlib.metadata
import re

project = "Virtue"
author = "Julian Berman"
copyright = "2014, " + author

release = importlib.metadata.version("virtue")
version = release.partition("-")[0]

language = "en"
default_role = "any"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.spelling",
]

pygments_style = "lovelace"
pygments_dark_style = "one-dark"

html_theme = "furo"

# = Builders =

nitpick_ignore = [
    # Missing from pyrsistent docs
    ("py:class", "pyrsistent.typing.PMap"),
    ("py:class", "pyrsistent.typing.PVector"),
]


def entire_domain(host):
    return r"http.?://" + re.escape(host) + r"($|/.*)"


linkcheck_ignore = [
    entire_domain("img.shields.io"),
    "https://github.com/Julian/Virtue/actions",
    "https://github.com/Julian/Virtue/workflows/CI/badge.svg",
]

# = Extensions =

# -- autodoc --

autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
}

# -- autosectionlabel --

autosectionlabel_prefix_document = True

# -- intersphinx --

intersphinx_mapping = dict(
    pyrsistent=("https://pyrsistent.readthedocs.io/en/latest/", None),
    python=("https://docs.python.org/3", None),
    twisted=("https://docs.twistedmatrix.com/en/stable/api/", None),
)

# -- sphinxcontrib-spelling --

spelling_word_list_filename = "spelling-wordlist.txt"
spelling_show_suggestions = True



"""
Runners execute loaded tests.
"""
import unittest
import warnings

import attr

from virtue.locators import ObjectLocator
from virtue.reporters import Counter


def run(tests=(), reporter=None, stop_after=None):
    """
    Run the tests that are loaded by each of the strings provided.

    Arguments:

        tests (collections.abc.Iterable):

            the collection of tests (specified as `str` s) to run

        reporter (twisted.trial.itrial.IReporter):

            a reporter to use for the run. If unprovided, the default is
            to return a `virtue.reporters.Counter` (which produces no
            output).

        stop_after (int):

            a number of non-successful tests to allow before stopping the run.
    """
    if reporter is None:
        reporter = Counter()
    if stop_after is not None:
        reporter = _StopAfterWrapper(reporter=reporter, limit=stop_after)

    locator = ObjectLocator()
    cases = (
        case
        for test in tests
        for loader in locator.locate_by_name(name=test)
        for case in loader.load()
    )
    suite = unittest.TestSuite(cases)
    getattr(reporter, "startTestRun", lambda: None)()
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        suite.run(reporter)
    getattr(reporter, "stopTestRun", lambda: None)()
    return reporter


@attr.s(eq=False)
class _StopAfterWrapper:
    """
    Wrap a reporter to stop after a specified number of non-successes.

    """

    _limit = attr.ib()
    _reporter = attr.ib()

    _seen = attr.ib(default=0)

    def __eq__(self, other):
        return self._reporter == other

    def __ne__(self, other):
        return not self == other

    def __getattr__(self, attr):
        return getattr(self._reporter, attr)

    def addError(self, *args, **kwargs):
        self._see_nonsuccess()
        self._reporter.addError(*args, **kwargs)

    def addFailure(self, *args, **kwargs):
        self._see_nonsuccess()
        self._reporter.addFailure(*args, **kwargs)

    def addUnexpectedSuccess(self, *args, **kwargs):
        self._see_nonsuccess()
        self._reporter.addUnexpectedSuccess(*args, **kwargs)

    def _see_nonsuccess(self):
        self._seen += 1
        if self._seen == self._limit:
            self.shouldStop = True



from textwrap import dedent

try:
    from pkgutil import resolve_name
except ImportError:
    from pkgutil_resolve_name import resolve_name  # type: ignore

import click
import twisted.trial.reporter

from virtue.reporters import ComponentizedReporter
from virtue.runner import run


class _Reporter(click.ParamType):
    name = "reporter"

    _BUILT_IN = {
        "bwverbose": twisted.trial.reporter.VerboseTextReporter,
        "default": ComponentizedReporter,
        "subunit": twisted.trial.reporter.SubunitReporter,
        "summary": twisted.trial.reporter.MinimalReporter,
        "text": twisted.trial.reporter.TextReporter,
        "timing": twisted.trial.reporter.TimingTextReporter,
        "tree": twisted.trial.reporter.TreeReporter,
        "verbose": twisted.trial.reporter.VerboseTextReporter,
    }

    def convert(self, value, param, ctx):
        if not isinstance(value, str):
            return value

        Reporter = self._BUILT_IN.get(value)
        if Reporter is None:
            try:
                Reporter = resolve_name(value)
            except (ValueError, ImportError, AttributeError):
                raise click.BadParameter(f"{value!r} is not a known reporter")

        return Reporter()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(prog_name="virtue")
@click.option(
    "--reporter",
    default="default",
    help=dedent(
        """
        the name of a reporter to use for outputting test results.
        Can either be a fully qualified object name (e.g.
        mypackage.MyReporter) or one of the builtin reporter names:
        """,
    )
    + ", ".join(_Reporter._BUILT_IN),
    type=_Reporter(),
)
@click.option(
    "-x",
    "--stop-after",
    default=None,
    count=True,
    help=(
        "stop the test run after unsuccessful results. May be "
        "repeated to stop after that many non-successes."
    ),
)
@click.argument("tests", nargs=-1)
@click.pass_context
def main(context, **kwargs):
    """
    `virtue` discovers and runs tests found in the given objects.

    Provide it with one or more tests (packages, modules or objects) to run.

    """
    result = run(**kwargs)
    context.exit(not result.testsRun or not result.wasSuccessful())



"""
Loaders take a named test case and load the test appropriately.
"""
from __future__ import annotations

from typing import TYPE_CHECKING
import itertools

from attrs import field, frozen

if TYPE_CHECKING:
    import twisted.python.modules

    from virtue.locators import ObjectLocator


@frozen
class AttributeLoader:
    """
    I load a test case by instantiating a class with a given attribute name.

    This is the typical way that `unittest.TestCase` methods are loaded:
    by calling ``TestCase("test_something")`` (and then by calling
    :meth:`~unittest.TestCase.run` on the resulting instance to run the
    selected test method).
    """

    cls: type
    attribute: str

    def load(self):
        """
        Load as a single test.
        """
        return [self.cls(self.attribute)]


@frozen
class ModuleLoader:
    """
    I load a test case by locating tests in the module with the given name.
    """

    locator: ObjectLocator = field(repr=False)
    module: twisted.python.modules.PythonModule

    def load(self):
        """
        Load all test cases in the module.
        """
        class_loaders = self.locator.locate_in_module(self.module.load())
        return itertools.chain.from_iterable(
            class_loader.load() for class_loader in class_loaders
        )



"""
Loaders find tests which are referenced by names, preparing them for running.
"""
from unittest import TestCase
import inspect

try:
    from pkgutil import resolve_name
except ImportError:
    from pkgutil_resolve_name import resolve_name  # type: ignore

from attrs import define, field
from twisted.python.modules import getModule as get_module
from twisted.python.reflect import fullyQualifiedName as fully_qualified_name
from twisted.trial.runner import filenameToModule as filename_to_module

from virtue.loaders import AttributeLoader, ModuleLoader


class UnableToLoad(Exception):
    """
    A test couldn't be loaded.
    """


def prefixed_by(prefix):
    """
    Make a callable returning True for names starting with the given prefix.

    The returned callable takes two arguments, the attribute or name of
    the object, and possibly its corresponding value (which is ignored),
    as suitable for use with `ObjectLocator.is_test_module` and
    `ObjectLocator.is_test_method`.

    """

    def prefixed_by_(name, value=None):
        return name.startswith(prefix)

    prefixed_by_.__name__ += prefix
    return prefixed_by_


def inherits_from_TestCase(attr, cls):
    """
    Return true if a class inherits from `unittest.TestCase`.
    """
    return issubclass(cls, TestCase)


@define
class ObjectLocator:
    """
    I locate test cases on an object: a package, module or test class.

    Arguments:

        is_test_method (collections.abc.Callable):

            decide whether the provided object is a test method
            or not. By default, callable objects whose names
            (``__name__``s) start with ``test_`` are considered
            test methods.

        is_test_class (collections.abc.Callable):

            decide whether the provided object is a test
            class or not. By default, objects inheriting from
            `unittest.TestCase` are considered test cases.

        is_test_module (collections.abc.Callable):

            decide whether the provided object is a test module
            or not. By default, modules whose names start with
            ``test_`` are considered to be test modules.

    """

    #: Whether an object is a test method or not
    is_test_method = field(default=prefixed_by("test"), repr=False)
    #: Whether an object is a test class or not
    is_test_class = field(default=inherits_from_TestCase, repr=False)
    #: Whether an object is a test module or not
    is_test_module = field(default=prefixed_by("test_"), repr=False)

    def __attrs_post_init__(self):
        is_cls, self.is_test_class = self.is_test_class, lambda attr, cls: (
            inspect.isclass(cls) and is_cls(attr, cls)
        )
        is_meth, self.is_test_method = self.is_test_method, lambda attr, val: (
            callable(val) and is_meth(attr, val)
        )

    def locate_by_name(self, name):
        """
        Locate any tests found in the object referred to by the given name.

        The name should be a fully qualified object name. (E.g., the fully
        qualified object name of this function is
        ``virtue.locators.ObjectLocator.locate_by_name``).

        A path may also alternatively used, but no `PYTHONPATH`
        modification will be done, so the file must be importable
        without modification.
        """
        try:
            obj = resolve_name(name)
        except ValueError:
            try:
                obj = filename_to_module(name)
            except ValueError as error:
                raise FileNotFoundError(error)

        try:
            return self.locate_in(obj)
        except UnableToLoad:
            class_name, _, method_name = name.rpartition(".")

            try:
                cls = resolve_name(class_name)
            except ValueError:
                class_name, _, method_name = name.rpartition(".")
                cls = resolve_name(class_name)
                if inspect.isclass(cls):
                    return [AttributeLoader(cls=cls, attribute=method_name)]
            else:
                if inspect.isclass(cls):
                    return [AttributeLoader(cls=cls, attribute=method_name)]

            # Aliased attributes
            fqon = fully_qualified_name(obj)
            class_name, _, method_name = fqon.rpartition(".")
            cls = resolve_name(class_name)
            if inspect.isclass(cls):
                return [AttributeLoader(cls=cls, attribute=method_name)]

            raise

    def locate_in(self, obj):
        """
        Attempt to locate the test cases in the given object (of any kind).
        """
        # We don't use inspect.getmembers because its predicate only
        # takes the value, not the attr name.
        if inspect.ismodule(obj):
            is_package = getattr(obj, "__path__", None)
            if is_package is not None:
                return self.locate_in_package(obj)
            return self.locate_in_module(obj)
        elif inspect.isclass(obj):
            return self.locate_in_class(obj)
        else:
            raise UnableToLoad(
                f"Can't determine the appropriate way to load {obj!r}",
            )

    def locate_in_package(self, package):
        """
        Locate all of the test cases contained in the given package.
        """
        for module in get_module(package.__name__).walkModules():
            _, _, name = module.name.rpartition(".")
            if self.is_test_module(name):
                yield ModuleLoader(locator=self, module=module)

    def locate_in_module(self, module):
        """
        Locate all of the test cases contained in the given module.
        """
        for attribute, value in inspect.getmembers(module):
            if self.is_test_class(attribute, value):
                yield from self.locate_in_class(value)

    def locate_in_class(self, cls):
        """
        Locate the methods on the given class that are test cases.
        """
        for attribute, value in inspect.getmembers(cls):
            if self.is_test_method(attribute, value):
                yield AttributeLoader(cls=cls, attribute=attribute)



from unittest import TestCase, expectedFailure


class Foo(TestCase):
    def test_foo(self):
        self.fail("Nope!")

    @expectedFailure
    def test_bar(self):
        pass

    @expectedFailure
    def test_baz(self):
        pass

    @expectedFailure
    def test_quux(self):
        pass

    def test_spam(self):
        pass



from unittest import TestCase


class Foo(TestCase):
    def test_a(self):
        self.skipTest("Skipped!")

    def test_b(self):
        self.skipTest("Skipped 2!")

    def test_c(self):
        self.skipTest("Skipped 2!")



from unittest import TestCase

from virtue import reporters


class TestRecorder(TestCase):
    def setUp(self):
        self.recorder = reporters.Recorder()

    def test_it_records_errors(self):
        error, exc_info = object(), object()
        self.recorder.addError(error, exc_info)
        self.assertEqual(self.recorder.errors, [(error, exc_info)])
        self.assertFalse(self.recorder.wasSuccessful())

    def test_it_records_failures(self):
        failure, exc_info = object(), object()
        self.recorder.addFailure(failure, exc_info)
        self.assertEqual(self.recorder.failures, [(failure, exc_info)])
        self.assertFalse(self.recorder.wasSuccessful())

    def test_it_records_successes(self):
        success = object()
        self.recorder.addSuccess(success)
        self.assertEqual(self.recorder.successes, [success])
        self.assertTrue(self.recorder.wasSuccessful())

    def test_it_records_unexpected_successes(self):
        success = object()
        self.recorder.addUnexpectedSuccess(success)
        self.assertEqual(self.recorder.unexpected_successes, [success])
        self.assertFalse(self.recorder.wasSuccessful())

    def test_testsRun(self):
        test, exc_info, reason = object(), object(), "Because!"
        self.recorder.addError(test=test, exc_info=exc_info)
        self.recorder.addFailure(test=test, exc_info=exc_info)
        self.recorder.addSkip(test=test, reason=reason)
        self.recorder.addSuccess(test=test)
        self.recorder.addUnexpectedSuccess(test=test)
        self.assertEqual(self.recorder.testsRun, 5)



from unittest import TestCase


class Foo(TestCase):
    def test_foo(self):
        pass

    def test_bar(self):
        self.fail("I fail.")


class Bar(TestCase):
    def test_foo(self):
        self.fail("I fail too.")

    def test_bar(self):
        pass



from unittest import TestCase


def create_test(name):
    methods = {f"test_{i}{j}": lambda _: None for j, i in enumerate(name)}
    return type(f"Test{name}", (TestCase,), methods)


TestFoo = create_test("Foo")



from unittest import TestCase

something = 12


class Foo(TestCase):
    def stuff(self):
        pass

    def test_foo(self):
        pass


class Bar:
    def stuff(self):
        pass

    def test_foo(self):
        pass


class Baz:
    def test_bar(self):
        pass

    def test_baz(self):
        pass


def test_123():
    pass



from operator import attrgetter
from unittest import TestCase
import shutil

from twisted.python.filepath import FilePath
from twisted.python.reflect import fullyQualifiedName

from virtue import locators
from virtue.loaders import AttributeLoader


class TestObjectLocator(TestCase):
    def test_it_finds_tests_in_an_object_specified_by_name(self):
        locator = locators.ObjectLocator()
        self.assertEqual(
            list(locator.locate_by_name(fullyQualifiedName(self.__class__))),
            list(locator.locate_in(self.__class__)),
        )

    def test_it_can_locate_methods_directly_by_name(self):
        locator = locators.ObjectLocator()
        this_name = "test_it_can_locate_methods_directly_by_name"
        this_fully_qualified_name = ".".join(
            [fullyQualifiedName(self.__class__), this_name],
        )
        self.assertEqual(
            list(locator.locate_by_name(this_fully_qualified_name)),
            [AttributeLoader(cls=self.__class__, attribute=this_name)],
        )

    def test_it_can_locate_aliased_methods_directly_by_name(self):
        locator = locators.ObjectLocator()
        this_name = "test_it_can_locate_aliased_methods_directly_by_name"
        aliased_fully_qualified_name = ".".join(
            [self.__class__.__module__, "aliased"],
        )
        self.assertEqual(
            list(locator.locate_by_name(aliased_fully_qualified_name)),
            [AttributeLoader(cls=self.__class__, attribute=this_name)],
        )

    def test_it_finds_methods_on_test_cases(self):
        locator = locators.ObjectLocator(
            is_test_method=locators.prefixed_by("TEST"),
        )

        class ASampleTestCase(TestCase):
            def not_a_test(self):
                pass

            def TEST1(self):
                pass

            def TEST_2(self):
                pass

        cases = locator.locate_in(ASampleTestCase)
        self.assertEqual(
            sorted(cases, key=attrgetter("attribute")),
            [
                AttributeLoader(cls=ASampleTestCase, attribute="TEST1"),
                AttributeLoader(cls=ASampleTestCase, attribute="TEST_2"),
            ],
        )

    def test_by_default_it_looks_for_methods_prefixed_by_test(self):
        locator = locators.ObjectLocator()

        class ASampleTestCase(TestCase):
            def not_a_test(self):
                pass

            def TEST1(self):
                pass

            def test_foo(self):
                pass

            def testBar(self):
                pass

        cases = locator.locate_in(ASampleTestCase)
        self.assertEqual(
            sorted(cases, key=attrgetter("attribute")),
            [
                AttributeLoader(cls=ASampleTestCase, attribute="testBar"),
                AttributeLoader(cls=ASampleTestCase, attribute="test_foo"),
            ],
        )

    def test_it_ignores_non_callable_test_methods(self):
        locator = locators.ObjectLocator()

        class ASampleTestCase(TestCase):
            test_foo = 12

            def test_bar(self):
                pass

        cases = locator.locate_in(ASampleTestCase)
        self.assertEqual(
            sorted(cases, key=attrgetter("attribute")),
            [
                AttributeLoader(cls=ASampleTestCase, attribute="test_bar"),
            ],
        )

    def test_it_loads_methods_from_dynamically_created_test_case_classes(self):
        locator = locators.ObjectLocator()
        from virtue.tests.samples.dynamic_test import TestFoo

        name = "virtue.tests.samples.dynamic_test.TestFoo.test_F0"
        self.assertEqual(
            list(locator.locate_by_name(name)),
            [AttributeLoader(cls=TestFoo, attribute="test_F0")],
        )

    def test_it_loads_methods_from_test_case_classes_with_mixin_methods(self):
        locator = locators.ObjectLocator()
        from virtue.tests.samples.mixin import TestFoo

        name = "virtue.tests.samples.mixin.TestFoo.test_baz"
        self.assertEqual(
            list(locator.locate_by_name(name)),
            [AttributeLoader(cls=TestFoo, attribute="test_baz")],
        )

    def test_it_finds_test_case_classes_on_modules(self):
        locator = locators.ObjectLocator(
            is_test_class=(lambda attr, value: attr != "Foo"),
        )
        from virtue.tests.samples import module_for_TestObjectLocator as module

        cases = locator.locate_in(module)
        self.assertEqual(
            sorted(cases, key=attrgetter("attribute")),
            [
                AttributeLoader(cls=module.Baz, attribute="test_bar"),
                AttributeLoader(cls=module.Baz, attribute="test_baz"),
                AttributeLoader(cls=module.Bar, attribute="test_foo"),
            ],
        )

    def test_by_default_it_looks_for_classes_inheriting_TestCase(self):
        locator = locators.ObjectLocator()
        from virtue.tests.samples import module_for_TestObjectLocator as module

        cases = locator.locate_in(module)
        self.assertEqual(
            sorted(cases, key=attrgetter("attribute")),
            [
                AttributeLoader(cls=module.Foo, attribute="test_foo"),
            ],
        )

    def test_it_finds_test_cases_recursively_in_packages(self):
        locator = locators.ObjectLocator(
            is_test_module=(lambda name: name != "foo"),
        )
        package = self.create_package_with_tests(locator)
        cases = locator.locate_in(package)
        self.assertEqual(
            sorted(case.module.name for case in cases),
            [
                "virtue.tests.temp",
                "virtue.tests.temp.bar",
                "virtue.tests.temp.sub",
                "virtue.tests.temp.sub.test_quux",
                "virtue.tests.temp.test_baz",
            ],
        )

    def test_by_default_it_finds_test_cases_in_modules_named_test_(self):
        locator = locators.ObjectLocator()
        package = self.create_package_with_tests(locator)
        cases = locator.locate_in(package)
        self.assertEqual(
            sorted(case.module.name for case in cases),
            [
                "virtue.tests.temp.sub.test_quux",
                "virtue.tests.temp.test_baz",
            ],
        )

    def test_it_knows_what_it_does_not_know(self):
        with self.assertRaises(locators.UnableToLoad):
            locators.ObjectLocator().locate_in(object())

    def create_package_with_tests(self, locator):
        package_path = FilePath(__file__).sibling(b"temp")
        package_path.makedirs()
        self.addCleanup(shutil.rmtree, package_path.path)

        package_path.child(b"__init__.py").setContent(b"")
        package_path.child(b"bar.py").setContent(b"")
        package_path.child(b"test_baz.py").setContent(b"")

        subpackage = package_path.child(b"sub")
        subpackage.makedirs()
        subpackage.child(b"__init__.py").setContent(b"")
        subpackage.child(b"test_quux.py").setContent(b"")

        from virtue.tests import temp as package

        return package


# Used to check that we locate or blow up properly on unbound methods
aliased_attr = "test_it_can_locate_aliased_methods_directly_by_name"
aliased = getattr(TestObjectLocator, aliased_attr)



"""
Outputting and reporting for virtuous test runs.
"""
from collections import defaultdict
from traceback import format_exception
import sys
import time

from pyrsistent import m, v
from pyrsistent.typing import PMap, PVector
import attrs


@attrs.frozen
class _DelayedMessage:
    """
    A test status message that will be shown at the end of a run.
    """

    width: int
    status: str
    subject: str
    body: str = ""

    def lines(self):
        yield "=" * self.width
        yield "\n"
        yield self.status
        yield "\n"
        if self.body:
            yield self.body


class Outputter:
    _last_test_class = None
    _last_test_module = None
    _current_subtests_test = None

    FAILED, PASSED = "FAILED", "PASSED"
    ERROR, FAIL, OK, SKIPPED = "[ERROR]", "[FAIL]", "[OK]", "[SKIPPED]"
    EXPECTED_FAILURE, UNEXPECTED_SUCCESS = "[XFAIL]", "[UNEXPECTED SUCCESS]"

    _COLORS = [
        ("_error", "RED", ERROR),
        ("_fail", "RED", FAIL),
        ("_failed", "RED", FAILED),
        ("_ok", "GREEN", OK),
        ("_passed", "GREEN", PASSED),
        ("_skipped", "BLUE", SKIPPED),
        ("_unexpected_success", "YELLOW", UNEXPECTED_SUCCESS),
        ("_expected_failure", "BLUE", EXPECTED_FAILURE),
    ]

    def __init__(self, colored=True, indent=" " * 2, line_width=120):
        self.indent = indent
        self.line_width = line_width

        if colored:
            from colorama import Fore, Style

            for attribute, color, text in self._COLORS:
                color = getattr(Fore, color)
                message = f"{Style.BRIGHT}{color}{text}{Style.RESET_ALL}"
                setattr(self, attribute, message)
        else:
            for attribute, _, text in self._COLORS:
                setattr(self, attribute, text)

        self._later = {
            status: defaultdict(list)
            for status in [
                self.SKIPPED,
                self.EXPECTED_FAILURE,
                self.FAIL,
                self.UNEXPECTED_SUCCESS,
                self.ERROR,
            ]
        }

    def _show_later(self, **kwargs):
        message = _DelayedMessage(width=self.line_width, **kwargs)
        self._later[message.status][message.body].append(message)

    def run_started(self):
        pass

    def run_stopped(self, recorder, runtime):
        for status in self._later.values():
            for messages in status.values():
                if not messages:
                    continue
                yield "\n"
                yield from messages[0].lines()
                for message in messages:
                    yield "\n"
                    yield message.subject.id()

        yield "\n"
        yield "-" * self.line_width

        count = recorder.testsRun
        tests = "tests" if count != 1 else "test"

        subcount = recorder.subtests
        if subcount:
            pluralized = "subtests" if subcount != 1 else "subtest"
            subtests = f" with {subcount} {pluralized}"
        else:
            subtests = ""

        yield f"\nRan {count} {tests}{subtests} in {runtime:.3f}s\n\n"
        if recorder.wasSuccessful():
            yield self._passed
        else:
            yield self._failed

        if recorder.testsRun:
            yield " ("
            summary = []
            for attribute in (
                "successes",
                "skips",
                "failures",
                "errors",
                "expected_failures",
                "unexpected_successes",
                "subtest_failures",
                "subtest_errors",
            ):
                subcount = len(getattr(recorder, attribute))
                if subcount:
                    summary.append(f"{attribute}={subcount}")
            yield ", ".join(summary)
            yield ")"

        yield "\n"

    def test_started(self, test):
        cls = test.__class__
        module = cls.__module__

        if self._last_test_module != module:
            self._last_test_module = module
            yield module
            yield "\n"

        if self._last_test_class != cls:
            self._last_test_class = cls
            yield self.indent
            yield cls.__name__
            yield "\n"

    def test_stopped(self, test):
        pass

    def test_errored(self, test, exc_info):
        self._show_later(
            status=self.ERROR,
            body="".join(format_exception(*exc_info)),
            subject=test,
        )
        return self.format_line(test, self._error)

    def test_failed(self, test, exc_info):
        self._show_later(
            status=self.FAIL,
            body="".join(format_exception(*exc_info)),
            subject=test,
        )
        return self.format_line(test, self._fail)

    def test_skipped(self, test, reason):
        self._show_later(status=self.SKIPPED, body=reason, subject=test)
        return self.format_line(test, self._skipped)

    def test_expectedly_failed(self, test, exc_info):
        self._show_later(status=self.EXPECTED_FAILURE, subject=test)
        return self.format_line(test, self._expected_failure)

    def test_unexpectedly_succeeded(self, test):
        self._show_later(status=self.UNEXPECTED_SUCCESS, subject=test)
        return self.format_line(test, self._unexpected_success)

    def test_succeeded(self, test):
        return self.format_line(test, self._ok)

    def subtest_succeeded(self, test, subtest):
        pass

    def subtest_failed(self, test, subtest, exc_info):
        self._show_later(
            status=self.FAIL,
            body="".join(format_exception(*exc_info)),
            subject=subtest,
        )
        return self.format_subtest_result(test, subtest, self._fail)

    def subtest_errored(self, test, subtest, exc_info):
        self._show_later(
            status=self.ERROR,
            body="".join(format_exception(*exc_info)),
            subject=subtest,
        )
        return self.format_subtest_result(test, subtest, self._error)

    def format_line(self, test, result):
        before = f"{self.indent}{self.indent}{test._testMethodName} ..."
        return self._pad_center(left=before, right=result) + "\n"

    def format_subtest_result(self, test, subtest, result):
        if self._current_subtests_test != test.id():
            before = f"{self.indent}{self.indent}{test._testMethodName}\n"
        else:
            before = ""
        self._current_subtests_test = test.id()
        line = f"{self.indent * 3}{subtest._subDescription()[1:-1]} ..."
        return f"{before}{self._pad_center(left=line, right=result)}\n"

    def _pad_center(self, left, right):
        space = self.line_width - len(left) - len(right)
        return left + " " * space + right


@attrs.define(slots=False)
class Counter:
    """
    A counter is a recorder that does not hold references to tests it sees.
    """

    errors: int = 0
    failures: int = 0
    expected_failures: int = 0
    unexpected_successes: int = 0
    successes: int = 0

    subtest_successes: int = 0
    subtest_failures: int = 0
    subtest_errors: int = 0

    shouldStop = False

    @property
    def count(self):
        return sum(attrs.astuple(self))

    testsRun = count

    def startTest(self, test):
        pass

    def stopTest(self, test):
        pass

    def addError(self, test, exc_info):
        self.errors += 1

    def addFailure(self, test, exc_info):
        self.failures += 1

    def addExpectedFailure(self, *args, **kwargs):
        self.expected_failures += 1

    def addUnexpectedSuccess(self, test):
        self.unexpected_successes += 1

    def addSuccess(self, test):
        self.successes += 1

    def addSubTest(self, test, subtest, outcome):
        if outcome is None:
            self.subtest_successes += 1
        elif issubclass(outcome[0], test.failureException):
            self.subtest_failures += 1
        else:
            self.subtest_errors += 1


@attrs.define(slots=False)
class Recorder:
    errors: PVector = v()
    failures: PVector = v()
    skips: PVector = v()
    successes: PVector = v()
    expected_failures: PVector = v()
    unexpected_successes: PVector = v()

    subtest_successes: PMap = m()
    subtest_failures: PMap = m()
    subtest_errors: PMap = m()

    shouldStop = False

    @property
    def testsRun(self):
        fields = attrs.astuple(
            self,
            filter=lambda f, _: not f.name.startswith("subtest_"),
        )
        # It seems addSuccess is called for tests with all passing subtests
        # but the reverse isn't true if a subtest fails...
        tests_with_subtests = len(self.subtest_failures | self.subtest_errors)
        return sum(len(each) for each in fields) + tests_with_subtests

    @property
    def subtests(self):
        return sum(
            1
            for each in (
                self.subtest_successes,
                self.subtest_failures,
                self.subtest_errors,
            )
            for value in each.values()
            for _ in value
        )

    def startTestRun(self):
        pass

    def stopTestRun(self):
        pass

    def startTest(self, test):
        pass

    def stopTest(self, test):
        pass

    def addError(self, test, exc_info):
        self.errors = self.errors.append((test, exc_info))

    def addFailure(self, test, exc_info):
        self.failures = self.failures.append((test, exc_info))

    def addExpectedFailure(self, test, exc_info):
        self.expected_failures = self.expected_failures.append(
            (test, exc_info),
        )

    def addSkip(self, test, reason):
        self.skips = self.skips.append(test)

    def addUnexpectedSuccess(self, test):
        self.unexpected_successes = self.unexpected_successes.append(test)

    def addSuccess(self, test):
        self.successes = self.successes.append(test)

    def addSubTest(self, test, subtest, outcome):
        if outcome is None:
            self.subtest_successes = self.subtest_successes.set(
                test,
                self.subtest_successes.get(test, v()).append(subtest),
            )
        elif issubclass(outcome[0], test.failureException):
            self.subtest_failures = self.subtest_failures.set(
                test,
                self.subtest_failures.get(test, v()).append(subtest),
            )
        else:
            self.subtest_errors = self.subtest_errors.set(
                test,
                self.subtest_errors.get(test, v()).append(subtest),
            )

    def wasSuccessful(self):
        return not (
            self.errors
            or self.failures
            or self.unexpected_successes
            or self.subtest_failures
            or self.subtest_errors
        )


@attrs.define(slots=False)
class ComponentizedReporter:
    outputter: Outputter = attrs.field(factory=Outputter)
    recorder = attrs.field(factory=Recorder, repr=False)
    stream = attrs.field(default=sys.stdout)
    _time = attrs.field(default=time.time, repr=False)

    failfast = False  # FIXME: needed for subtests?
    shouldStop = False

    @property
    def testsRun(self):
        return self.recorder.testsRun

    def startTestRun(self):
        self._start_time = self._time()
        self.recorder.startTestRun()
        self.stream.writelines(self.outputter.run_started() or "")

    def stopTestRun(self):
        self.recorder.stopTestRun()
        runtime = self._time() - self._start_time
        self.stream.writelines(
            self.outputter.run_stopped(self.recorder, runtime) or "",
        )

    def startTest(self, test):
        self.recorder.startTest(test)
        self.stream.writelines(self.outputter.test_started(test) or "")

    def stopTest(self, test):
        self.recorder.stopTest(test)
        self.stream.writelines(self.outputter.test_stopped(test) or "")

    def addError(self, test, exc_info):
        self.recorder.addError(test, exc_info)
        self.stream.writelines(
            self.outputter.test_errored(test, exc_info) or "",
        )

    def addFailure(self, test, exc_info):
        self.recorder.addFailure(test, exc_info)
        self.stream.writelines(
            self.outputter.test_failed(test, exc_info) or "",
        )

    def addSkip(self, test, reason):
        self.recorder.addSkip(test, reason)
        self.stream.writelines(self.outputter.test_skipped(test, reason) or "")

    def addExpectedFailure(self, test, exc_info):
        self.recorder.addExpectedFailure(test, exc_info)
        self.stream.writelines(
            self.outputter.test_expectedly_failed(test, exc_info) or "",
        )

    def addUnexpectedSuccess(self, test):
        self.recorder.addUnexpectedSuccess(test)
        self.stream.writelines(
            self.outputter.test_unexpectedly_succeeded(test) or "",
        )

    def addSuccess(self, test):
        self.recorder.addSuccess(test)
        self.stream.writelines(self.outputter.test_succeeded(test) or "")

    def addSubTest(self, test, subtest, outcome):
        self.recorder.addSubTest(test, subtest, outcome)
        if outcome is None:
            output = self.outputter.subtest_succeeded(test, subtest)
        elif issubclass(outcome[0], test.failureException):
            output = self.outputter.subtest_failed(test, subtest, outcome)
        else:
            output = self.outputter.subtest_errored(test, subtest, outcome)
        self.stream.writelines(output or "")

    def wasSuccessful(self):
        return self.recorder.wasSuccessful()



from unittest import TestCase

from twisted.python.modules import getModule as get_module

from virtue import loaders, locators


class TestAttributeLoader(TestCase):
    def test_it_loads_attributes(self):
        cls, attr = self.__class__, "test_it_loads_attributes"
        loader = loaders.AttributeLoader(cls=cls, attribute=attr)
        self.assertEqual(list(loader.load()), [cls(attr)])

    def test_eq_neq(self):
        cls = self.__class__
        loader = loaders.AttributeLoader(cls=cls, attribute="test_eq")
        self.assertTrue(
            loader == loaders.AttributeLoader(cls=cls, attribute="test_eq"),
        )
        self.assertFalse(
            loader != loaders.AttributeLoader(cls=cls, attribute="test_eq"),
        )
        self.assertFalse(
            loader == loaders.AttributeLoader(cls=cls, attribute="test_neq"),
        )
        self.assertTrue(
            loader != loaders.AttributeLoader(cls=cls, attribute="test_neq"),
        )

    def test_repr(self):
        loader = loaders.AttributeLoader(
            cls=self.__class__,
            attribute="test_repr",
        )
        self.assertEqual(
            repr(loader),
            "AttributeLoader(cls=<class 'virtue.tests.test_loaders.TestAttributeLoader'>, attribute='test_repr')",  # noqa: E501
        )


class TestModuleLoader(TestCase):
    locator = locators.ObjectLocator()

    def test_it_loads_modules(self):
        module = get_module("virtue.tests.samples.module_for_TestLoaders")
        loader = loaders.ModuleLoader(locator=self.locator, module=module)

        cases = module.load()
        self.assertEqual(
            list(loader.load()),
            [
                cases.Baz("test_bar"),
                cases.Baz("test_baz"),
                cases.Foo("test_foo"),
            ],
        )

    def test_eq_neq(self):
        virtue, os = get_module("virtue"), get_module("os")
        loader = loaders.ModuleLoader(locator=self.locator, module=virtue)
        self.assertEqual(
            loader,
            loaders.ModuleLoader(locator=self.locator, module=virtue),
        )
        self.assertNotEqual(
            loader,
            loaders.ModuleLoader(locator=self.locator, module=os),
        )

    def test_repr(self):
        virtue = get_module("virtue")
        loader = loaders.ModuleLoader(locator=self.locator, module=virtue)
        self.assertEqual(
            repr(loader),
            "ModuleLoader(module=PythonModule<'virtue'>)",
        )



from unittest import TestCase


class Foo(TestCase):
    def test_foo(self):
        self.fail("Nope!")

    def test_bar(self):
        raise Exception("Boom!")

    def test_baz(self):
        raise Exception("Whiz!")

    def test_quux(self):
        raise Exception("Bang!")

    def test_spam(self):
        raise Exception("Hiss!")



from unittest import TestCase
import os

from twisted.trial.reporter import TreeReporter

from virtue import _cli


def DumbReporter():
    return 12


class TestParser(TestCase):
    def parse_args(self, argv):
        return _cli.main.make_context("virtue", argv).params

    def test_it_parses_out_tests(self):
        arguments = self.parse_args(["foo", "bar", "baz"])
        self.assertEqual(list(arguments["tests"]), ["foo", "bar", "baz"])

    def test_it_retrieves_built_in_reporters_by_name(self):
        arguments = self.parse_args(["--reporter", "tree", "foo"])
        self.assertIsInstance(arguments["reporter"], TreeReporter)

    def test_it_retrieves_other_reporters_by_fully_qualified_name(self):
        arguments = self.parse_args(
            ["--reporter", "virtue.tests.test_cli.DumbReporter", "abc"],
        )
        self.assertEqual(arguments["reporter"], DumbReporter())

    def test_stop_after(self):
        arguments = self.parse_args(["-xxx", "bar", "baz"])
        self.assertEqual(
            (arguments["stop_after"], list(arguments["tests"])),
            (3, ["bar", "baz"]),
        )

    def test_stop_after_default(self):
        arguments = self.parse_args(["-x", "bar", "baz"])
        self.assertEqual(
            (arguments["stop_after"], list(arguments["tests"])),
            (1, ["bar", "baz"]),
        )


class TestMain(TestCase):
    # TODO: these write to stdout
    def test_it_exits_successfully_for_successful_runs(self):
        with self.assertRaises(SystemExit) as e:
            _cli.main(
                [
                    "--reporter",
                    "summary",
                    "virtue.tests.samples.one_successful_test",
                ],
            )
        self.assertEqual(e.exception.code, os.EX_OK)

    def test_it_exits_unsuccessfully_for_unsuccessful_runs(self):
        with self.assertRaises(SystemExit) as e:
            _cli.main(
                [
                    "--reporter",
                    "text",
                    "virtue.tests.samples.one_unsuccessful_test",
                ],
            )
        self.assertNotEqual(e.exception.code, os.EX_OK)

    def test_it_exits_unsuccessfully_for_unknown_reporters(self):
        with self.assertRaises(SystemExit) as e:
            _cli.main(
                [
                    "--reporter",
                    "non-existent reporter",
                    "virtue.tests.samples.one_unsuccessful_test",
                ],
            )
        self.assertNotEqual(e.exception.code, os.EX_OK)

    def test_it_exits_unsuccessfully_when_no_tests_ran(self):
        with self.assertRaises(SystemExit) as e:
            _cli.main(["virtue.tests.samples.no_tests"])
        self.assertNotEqual(e.exception.code, os.EX_OK)



from unittest import TestCase


class Foo(TestCase):
    def test_no_subtests(self):
        pass

    def test_subtests_one_fail_one_error(self):
        for i in range(4):
            with self.subTest(i=i):
                if i == 1:
                    self.fail("Fail!")
                elif i == 3:
                    raise ZeroDivisionError()


class Baz(TestCase):
    def test_passing_nested_subtest(self):
        with self.subTest(quux="hello", spam="eggs"):
            with self.subTest(i=3):
                pass

    def test_passing_subtest(self):
        with self.subTest(eggs=3):
            pass



from unittest import TestCase


class Foo(TestCase):
    def stuff(self):
        pass

    def test_foo(self):
        pass


class Baz(TestCase):
    def test_bar(self):
        pass

    def test_baz(self):
        pass
